/*
   I want to say a few words about the code:
   We have three essential libraries:
    - ESP8266Webserver wich will let us host a webserver on the NodeMCU
    - FastLED to control the leds. I prefer it over Adafruit Neopixel because it is faster and has more funtionality
    - SD.h for the reading and writing from the SD card

    We are hosting a webserver on the NodeMCU that communicates with the client (the website you see in your browser) over http POST und GET requests.
    These requests are handled in the "server handles" section
    Most of the FastLED operations are in the "FAST LED" section

*/

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <SPI.h>
#include <SD.h>
#include <FastLED.h>

#define STASSID "MY"         //SSID of your WIFI
#define STAPSK "d62m96f04xy!"      //password of your WIFI
#define DATA_PIN 3                  //pin of the LED strip
#define NUM_LEDS 256                //16*16 matrix = 256 leds

const char* ssid = STASSID;
const char* password = STAPSK;

ESP8266WebServer server(80);        //start the webserver on port 80 (standard html port)

File html;                          //file to read the html and push it to the client
File patterns;                      //file to read one of the patterns stored on the sd

CRGB leds[NUM_LEDS];                //initalise the array of the leds that we can later assign colorvalues
extern const uint8_t gamma8[];      //this is a forward declaration for the gamma correction (used to make the colors look more natural)
long pattern[256];                  //array of the current displayed led pattern
bool live = true;                   //determines if NodeMCU is in live draw mode or shows the saved patterns in a loop
long currentMillis = 0;             //we all know these variables.. blink without delay...
long prevMillis = millis();
int interval = 7000;
int loopIndex = 0;                  //index of the loop for displaying the saved images

/*------------------------------------------------------------
                      FAST LED
  -----------------------------------------------------------*/

void clearMatrix() {                                              //set all leds to black
  for (int i = 0; i < 256; i++) {
    leds[i] = CRGB(0, 0, 0);
  }
  FastLED.show();
}

void draw(int index, long color) {                                //draw one pixel on the matrix
  leds[index].setRGB(pgm_read_byte(&gamma8[(color >> 16) & 255]), //gamma correction. pgm_read_byte because the correction array is stored in flash
                     pgm_read_byte(&gamma8[(color >> 8) & 255]),
                     pgm_read_byte(&gamma8[(color) & 255]));
  FastLED.show();
}

void showPattern() {                                              //display pattern currently stored in pattern[256]
  for (int i = 0; i < 256; i++) {
    leds[i] = pattern[i];
  }
  FastLED.show();
}

/*------------------------------------------------------------
                      server handles
  -----------------------------------------------------------*/
void handleRoot() {                                               //handle normal opening of the website
  html = SD.open("index.html");
  server.streamFile(html, "text/html");                           //stream the contents of the .html file to the client to be displayed
  html.close();                                                   //close the file
}

void handleNotFound() {
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  server.send(404, "text/plain", message);
}

void handlePlain() {                                               //handles the post requests for clearing the canvas and drawing a pixel
  if (server.method() != HTTP_POST) {
    server.send(405, "text/plain", "Method Not Allowed");
  } else {
    server.send(200);

    if (server.hasArg("cl")) {                                    //cl for clear
      clearMatrix();
    } else {
      draw(server.arg("i").toInt(),                               //i for the index, c for the color
           strtoul(server.arg("c").c_str(), NULL, 16));    
    }
  }
}

void handleGet() {                                                //handles the request for getting the names of the stored patterns
  String names = getNames();
  server.send(200, "text/plain", names);
}

void handleLoad() {                                               //handles loading of a pattern
  String jsonarray = "[\"";                                       //response to the request will be a JSON array of the color values

  File loadPattern = SD.open(server.arg("id") + ".txt", FILE_READ);
  if (pattern) {
    int i = 0;                                                    //index how many values have been written to the array, used to exclude the final newline in the .txt file

    while (loadPattern.available()) {
      char inChar = loadPattern.read();                           //read a new char
      byte inCharAscii = inChar;                                  //save the ascii of the char
      if (inCharAscii == 10) {                                    //ASCII 10 = newline
        if (i < 255) {
          jsonarray += "\",\"";
          i++;
        }
      }
      else if (inCharAscii == 13) {                               //ASCII 13 = character return
      }
      else {
        jsonarray += inChar;                                      //schreiben des chars in das tempArray

      }
    }
    jsonarray += "\"]";                                           //close the JSON array
    loadPattern.close();                                          //close the file                                
    server.send(200, "text/plain", jsonarray);                    //send the response
  } else {                                                        
    Serial.println("Cant open names file"); 
    server.send(404, "text/plain", "error");
  }

  readFile(server.arg("id").toInt(), pattern);                    //show the loaded pattern
  showPattern();
}

void handleSave() {
  server.send(200, "text/plain", "done");                         //some cheating, not really done but we can still work with the request :)
  int i = 0;                                                      //index variable

  while (SD.exists(String(i) + ".txt")) {                         //iterate over the existing files till we have a free spot
    i++;
  }

  File newFile = SD.open(String(i) + ".txt", FILE_WRITE);         //create a new file for the new pattern
  newFile.print(server.arg("pt"));                                //pt for pattern, argument has the right format so we can just write it to SD
  newFile.close();

  File names = SD.open("names.txt", FILE_WRITE);                  
  if (names) {
    names.println(server.arg("name"));                            //add the newly added pattern to the names file
    names.close();
  } else {
    Serial.println("Can't open name.txt");
  }
  names.close();
}

void handleDelete() {                                             //response to the client
  server.send(200, "text/plain", "ok");

  long pos = findPos(server.arg("name").c_str());                 //find the position of the name to delete in the name file
  File names = SD.open("names.txt", FILE_WRITE);
  names.seek(pos);                                                //open the file with the cursor at the pos
  names.write('^');                                               //^ is the character for deleted names
  for (int i = 0; i < server.arg("name").length(); i++) {
    names.write(" ");                                             //there is no easy way to really delete something from a file so we just ust spaces
  }
  names.close();
}

void handleMode() {                                               //displaying saved patterns in a loop or draw live 
  server.send(200);
  if (server.arg("mode") == "true") {                             //toggle the variable
    live = true;
  }
  else {
    live = false;
  }
}

/*------------------------------------------------------------
                      file handling
  -----------------------------------------------------------*/

void parsePattern(long (&colorArray)[256]) {
  char buf[9];                                      //array zum Speichern eines Hexwertes (temp)
  int i = 0;                                        //Index für das buf array
  int j = 0;                                        //index für das patternArray
  while (patterns.available()) {
    char inChar = patterns.read();                  //ein neuen char einlesen
    byte inCharAscii = inChar;                      //asciiwert des chars speichern
    if (inCharAscii == 10) {                        //ASCII 10 = newline
      buf[i] = '\0';                                //Cstring terminieren
      colorArray[j] = strtoul(buf + 2, NULL, 16);   //cstring in unsigned long umwandeln
      j++;
      i = 0;
    }
    else if (inCharAscii == 13) {                   //ASCII 13 = character return
    }
    else {
      buf[i] = inChar;                              //add the inChar into a buffer
      i++;
    }
  }
}

void readFile(int index, long (&colorArray)[256]) { 
  patterns = SD.open((String)index + ".txt");       //open the file with the given index
  if (patterns) {                                   //opening succesful
    parsePattern(colorArray);                       //read the file and write to pattern array
  }
  else {
    Serial.println("Error opening file");
  }
  patterns.close();
}

String getNames() {                                 //get the names of the patterns
  String jsonarray = "[\"";                         //names will be send in a JSON array
  File names = SD.open("names.txt");
  if (names) {
    while (names.available()) {
      char inChar = names.read();                   //read a new char
      byte inCharAscii = inChar;                    //save the ASCII of the char
      if (inCharAscii == 10) {                      //ASCII 10 = newline
        jsonarray += "\",\"";
      }
      else if (inCharAscii == 13) {                 //ASCII 13 = character return
      }
      else {
        jsonarray += inChar;                        //write the char into the array
      }
    }
    jsonarray += "\"]";
    names.close();
    return jsonarray;
  } else {
    Serial.println("Cant open names file");
    return jsonarray;
  }
}

long findPos(const char searchString[20]) {         //finds the position of a string in a file
  File names = SD.open("names.txt");
  bool found = true;
  long pos = -1;
  int i = 0;                                        //Index für das buf array

  while (names.available()) {
    char inChar = names.read();                     
    byte inCharAscii = inChar; 
    if (inCharAscii == 10) {                        //ASCII 10 = newline
      if (found == true) {                          //if the characters of one line match the search string
        Serial.println("string found");
        return pos;
      }
      i = 0;
      found = true;
      pos = names.position();                       //save the position of the beginning of the line
    }
    else if (inCharAscii == 13) {                   //ASCII 13 = carriage return
    }
    else {
      if (found == true) {
        if (inChar == searchString[i]) {            
          found = true;
        }
        else {
          found = false;
        }
        i++;
      }
    }
  }
  return 1;
}

/*------------------------------------------------------------
                      setup
  -----------------------------------------------------------*/

void setup(void) {

  //begin wifi and serial
  Serial.begin(115200);                             
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  Serial.println("");

  // Wait for connection
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());

  if (MDNS.begin("esp8266")) {
    Serial.println("MDNS responder started");
  }

  //initialise the SD card on pin 4
  if (!SD.begin(4)) {                         
    Serial.println("initialization failed!");
    while (1);
  }
  Serial.println("initialization done.");

  /*
   * This sets wich function will be executed if a request is made on the specified url
   */
  server.on("/", handleRoot);
  server.on("/live", handlePlain);
  server.on("/get", handleGet);
  server.on("/save", handleSave);
  server.on("/delete", handleDelete);
  server.on("/load", handleLoad);
  server.on("/mode", handleMode);
  server.onNotFound(handleNotFound);

  //start the server
  server.begin();
  Serial.println("HTTP server started");

  //initialise FastLED
  FastLED.addLeds<WS2813, DATA_PIN, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(50);
  clearMatrix();
}

/*------------------------------------------------------------
                      looooooooooop
  -----------------------------------------------------------*/

void loop(void) {
  server.handleClient();  
  MDNS.update();

  currentMillis = millis();

  if (currentMillis - prevMillis >= interval && live == false) {  //if the mode is looping through the saved pictures and its time for a new one
    if (!(SD.exists(String(loopIndex) + ".txt"))) {               //check if file with given index exists,otherwise begin from beginning
      loopIndex = 0;
    }
    readFile(loopIndex, pattern);
    showPattern();
    prevMillis = millis(); 
    loopIndex++;
  }
}

const uint8_t PROGMEM gamma8[] = {                                                    //table for gamma correction
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,
  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,
  2,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,  4,  4,  5,  5,  5,
  5,  6,  6,  6,  6,  7,  7,  7,  7,  8,  8,  8,  9,  9,  9, 10,
  10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16,
  17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22, 23, 24, 24, 25,
  25, 26, 27, 27, 28, 29, 29, 30, 31, 32, 32, 33, 34, 35, 35, 36,
  37, 38, 39, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 50,
  51, 52, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 66, 67, 68,
  69, 70, 72, 73, 74, 75, 77, 78, 79, 81, 82, 83, 85, 86, 87, 89,
  90, 92, 93, 95, 96, 98, 99, 101, 102, 104, 105, 107, 109, 110, 112, 114,
  115, 117, 119, 120, 122, 124, 126, 127, 129, 131, 133, 135, 137, 138, 140, 142,
  144, 146, 148, 150, 152, 154, 156, 158, 160, 162, 164, 167, 169, 171, 173, 175,
  177, 180, 182, 184, 186, 189, 191, 193, 196, 198, 200, 203, 205, 208, 210, 213,
  215, 218, 220, 223, 225, 228, 231, 233, 236, 239, 241, 244, 247, 249, 252, 255
};
